EMAIL_EXTRACTION_PROMPT = """
You are a friendly email agent. Your job is to extract structured information from emails in accordance with the schema provided.
"""

AGENT_INSTRUCTIONS = """
## PERSISTENCE
You are an agent - please keep going until the user's query is completely
resolved, before ending your turn and yielding back to the user. Only
terminate your turn when you are sure that the problem is solved.

## TOOL CALLING
If you are not sure about file content or codebase structure pertaining to
the user's request, use your tools to read files and gather the relevant
information: do NOT guess or make up an answer.

## PLANNING
You MUST plan extensively before each function call, and reflect
extensively on the outcomes of the previous function calls. DO NOT do this
entire process by making function calls only, as this can impair your
ability to solve the problem and think insightfully.
"""



from yougotmail import YouGotMail
import os
from dotenv import load_dotenv
import json

load_dotenv()

ygm = YouGotMail(
    client_id=os.environ.get("MS_CLIENT_ID"),
    client_secret=os.environ.get("MS_CLIENT_SECRET"),
    tenant_id=os.environ.get("MS_TENANT_ID"),
    mongo_url=os.environ.get("MONGO_URL"),
    mongo_db_name=os.environ.get("MONGO_DB_NAME"),
    email_collection=os.environ.get("EMAIL_COLLECTION"),
    conversation_collection=os.environ.get("CONVERSATION_COLLECTION"),
    attachment_collection=os.environ.get("ATTACHMENT_COLLECTION"),
    aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
    aws_secret_access_key=os.environ.get("AWS_SECRET_ACCESS_KEY"),
    region_name=os.environ.get("REGION_NAME"),
    bucket_name=os.environ.get("BUCKET_NAME")
)

def test_retrieving_emails_with_storage():
    try:
        emails = ygm.get_emails(
            inbox=[os.environ.get("INBOX_1")],
            range="last_8_hours",
            storage="emails_and_attachments"
        )
        print(json.dumps(emails, indent=4))
    except Exception as e:
        print(f"Error: {e}")


from yougotmail.send.send import Send


TOOLS = [
    {
        "type": "function",
        "name": "get_weather",
        "description": "Get current temperature for a given location.",
        "strict": True,
        "parameters": {
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "description": "City and country e.g. Bogotá, Colombia",
                }
            },
            "required": ["location"],
            "additionalProperties": False,
        },
    }
]


class AI_TOOLS:
    def get_weather(self, location):
        return "Weather retrieved"



from yougotmail import YouGotMail
import os
from dotenv import load_dotenv
import json

load_dotenv()

ygm = YouGotMail(
    client_id=os.environ.get("MS_CLIENT_ID"),
    client_secret=os.environ.get("MS_CLIENT_SECRET"),
    tenant_id=os.environ.get("MS_TENANT_ID"),
    open_ai_api_key=os.environ.get("OPENAI_API_KEY"),
)


def test_ai_parsing():
    try:
        emails = ygm.ai_get_emails_with_structured_output(
            inbox=[os.environ.get("INBOX_1"), os.environ.get("INBOX_2")],
            range="last_8_hours",
            attachments=False,
            schema={
                "topic": {"type": "string", "description": "The topic of the email"},
                "sentiment": {
                    "type": "string",
                    "description": "what was the mood of the email and the sender writing it",
                },
            },
        )
        with open("emails_structured_output.json", "w") as f:
            json.dump(emails, f, indent=4)
    except Exception as e:
        print(f"Error: {e}")



from yougotmail import YouGotMail
import os
from dotenv import load_dotenv

load_dotenv()

ygm = YouGotMail(
    client_id=os.environ.get("MS_CLIENT_ID"),
    client_secret=os.environ.get("MS_CLIENT_SECRET"),
    tenant_id=os.environ.get("MS_TENANT_ID"),
)

def test_draft_email():
    try:
        draft_email = ygm.draft_email(
            inbox=os.environ.get("INBOX_1"),
            subject="test",
            importance="",
            email_body="<html><body><h1>Test Email</h1><p>This is a test email sent from YouGotMail.</p></body></html>",
            to_recipients=[os.environ.get("INBOX_1")],
            cc_recipients=[],
            bcc_recipients=[],
            attachments=[],
        )
        print(draft_email)
    except Exception as e:
        print(f"Error: {e}")

def test_sending_emails():
    try:
        send_email = ygm.send_email(
            inbox=os.environ.get("INBOX_1"),
            subject="test",
            importance="",
            email_body="<html><body><h1>Test Email</h1><p>This is a test email sent from YouGotMail.</p></body></html>",
            to_recipients=[os.environ.get("INBOX_1")],
            cc_recipients=[],
            bcc_recipients=[],
            attachments=[ # You can pass a URL (make sure it's accessible) or a local file path
                "https://www.google.com/images/branding/googlelogo/1x/googlelogo_color_272x92dp.png"
            ]
        )
        print(send_email)
    except Exception as e:
        print(f"Error: {e}")

def test_replying_to_email():
    try:
        reply_to_email = ygm.reply_to_email(
            inbox=os.environ.get("INBOX_1"),
            email_id=os.environ.get("EMAIL_ID"),
            email_body="This is a test reply to the email - AI signature",
            cc_recipients=["ai@delosone.com"]
        )
        print(reply_to_email)
    except Exception as e:
        print(f"Error: {e}")



import base64
from yougotmail._utils._utils import Utils

class Storage:
    def __init__(
        self,
        mongo_url="",
        mongo_db_name="",
        email_collection="",
        conversation_collection="",
        attachment_collection="",
        aws_access_key_id="",
        aws_secret_access_key="",
        region_name="",
        bucket_name="",
    ):
        # Store configuration but don't initialize connections
        self.mongo_config = {
            "url": mongo_url,
            "db_name": mongo_db_name,
            "email_collection": email_collection,
            "conversation_collection": conversation_collection,
            "attachment_collection": attachment_collection,
        }

        self.aws_config = {
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
            "region_name": region_name,
            "bucket_name": bucket_name,
        }

        self.utils = Utils()
        self._mongo_client = None
        self._s3_client = None
        self._db = None
        self._emails_collection = None
        self._conversations_collection = None
        self._attachments_collection = None
        self.enabled = self._check_credentials()
        if self.enabled:
            self._initialize_connections()

    def _check_credentials(self):
        # Check if we're in Lambda (only need region and bucket)
        lambda_environment = all(
            [
                self.aws_config["region_name"],
                self.aws_config["bucket_name"],
            ]
        )

        if all(self.mongo_config.values()) and (
            all(self.aws_config.values()) or lambda_environment
        ):
            return True
        else:
            return False

    def _initialize_connections(self):
        """Initialize MongoDB and S3 connections if credentials are available."""
        try:
            # Initialize MongoDB connection
            import pymongo
            self._mongo_client = pymongo.MongoClient(self.mongo_config["url"])
            self._db = self._mongo_client[self.mongo_config["db_name"]]
            self._emails_collection = self._db[self.mongo_config["email_collection"]]
            self._conversations_collection = self._db[self.mongo_config["conversation_collection"]]
            self._attachments_collection = self._db[self.mongo_config["attachment_collection"]]
            
            # Initialize S3 connection
            import boto3
            if all([self.aws_config["aws_access_key_id"], self.aws_config["aws_secret_access_key"]]):
                # If we have explicit credentials
                self._s3_client = boto3.client(
                    's3',
                    aws_access_key_id=self.aws_config["aws_access_key_id"],
                    aws_secret_access_key=self.aws_config["aws_secret_access_key"],
                    region_name=self.aws_config["region_name"]
                )
            else:
                # If we're in Lambda or using IAM roles
                self._s3_client = boto3.client(
                    's3',
                    region_name=self.aws_config["region_name"]
                )
        except Exception as e:
            print(f"Error initializing connections: {e}")
            self.enabled = False

    def store_emails(self, inbox_list):
        if not self.enabled:
            print("Storage is not configured - skipping email storage")
            return
        try:
            for inbox in inbox_list:
                for email in inbox["emails"]:
                    if self._check_if_email_exists(email["email_id"]):
                        print(
                            f"Email document already exists in mongo: {email['email_id']}"
                        )
                        continue
                    email["received_date"] = self.utils._format_date(
                        email["received_date"]
                    )
                    attachments = email["attachments"]
                    new_attachment_list = []
                    for attachment in attachments:
                        new_attachment = {
                            "attachment_id": attachment["attachment_id"],
                            "file_name": attachment["file_name"],
                            "date": self.utils._format_date(attachment["date"]),
                            "contentType": attachment["contentType"],
                        }
                        new_attachment_list.append(new_attachment)
                    email["attachments"] = new_attachment_list
                    self._emails_collection.insert_one(email)
                    print(f"Email document saved in mongo: {email['email_id']}")
        except Exception as e:
            print(f"Error in store_emails: {e}")
            return None

    def store_emails_and_attachments(self, inbox_list):
        if not self.enabled:
            print("Storage is not configured - skipping email storage")
            return
        try:
            for inbox in inbox_list:
                inbox_name = inbox["inbox"]
                emails = inbox["emails"]
                for email in emails:
                    if self._check_if_email_exists(email["email_id"]):
                        print(
                            f"Email document already exists in mongo: {email['email_id']}"
                        )
                        continue
                    email["received_date"] = self.utils._format_date(
                        email["received_date"]
                    )
                    attachments = email["attachments"]
                    if attachments:
                        new_attachment_list = []
                        for attachment in attachments:
                            file_metadata = self._store_attachments_in_s3(
                                inbox_name, email["email_id"], attachment
                            )
                            new_attachment = {
                                "attachment_id": attachment["attachment_id"],
                                "file_name": attachment["file_name"],
                                "date": self.utils._format_date(attachment["date"]),
                                "contentType": attachment["contentType"],
                                "url": file_metadata["url"],
                            }
                            new_attachment_list.append(new_attachment)
                        email["attachments"] = new_attachment_list
                    self._emails_collection.insert_one(email)
                    print(f"Email document saved in mongo: {email['email_id']}")
        except Exception as e:
            print(f"Error in store_emails: {e}")
            return None

    def store_attachments(self, attachments_list):
        if not self.enabled:
            print("Storage is not configured - skipping attachment storage")
            return
        for inbox in attachments_list:
            for attachment in inbox["attachments"]:
                if self._check_if_attachment_exists(attachment["attachment_id"]):
                    print(
                        f"Attachment document already exists in mongo: {attachment['attachment_id']}"
                    )
                    continue
                file_metadata = self._store_attachments_in_s3(
                    inbox["inbox"], attachment
                )
                new_attachment = {
                    "attachment_id": attachment["attachment_id"],
                    "file_name": attachment["file_name"],
                    "date": self.utils._format_date(attachment["date"]),
                    "contentType": attachment["contentType"],
                    "url": file_metadata["url"],
                }
                self._attachments_collection.insert_one(new_attachment)
                print(
                    f"Attachment document saved in mongo: {attachment['attachment_id']}"
                )

    def _check_if_email_exists(self, email_id):
        if not self.enabled:
            return False
        existing_email = self._emails_collection.find_one({"email_id": email_id})
        if existing_email:
            return True
        else:
            return False

    def _check_if_attachment_exists(self, attachment_id):
        if not self.enabled:
            return False
        existing_attachment = self._attachments_collection.find_one(
            {"attachment_id": attachment_id}
        )
        if existing_attachment:
            return True
        else:
            return False

    def _check_if_conversation_exists(self, conversation_id):
        if not self.enabled:
            return False
        existing_conversation = self._conversations_collection.find_one(
            {"conversation_id": conversation_id}
        )
        if existing_conversation:
            return True
        else:
            return False

    def _store_attachments_in_s3(self, inbox_name, email_id, attachment):
        if not self.enabled:
            return None
        try:
            attachment_id = attachment["attachment_id"]
            file_name = attachment["file_name"]
            file_extension = file_name.split(".")[-1]
            file_date = attachment["date"]
            file_name_with_underscores = attachment["file_name"].replace(" ", "_")
            contentType = ""
            body = ""

            if file_extension == "pdf":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "application/pdf"
            elif file_extension == "xlsx":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = (
                    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
                )
            elif file_extension == "docx":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
            elif file_extension == "pptx":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "application/vnd.openxmlformats-officedocument.presentationml.presentation"
            elif file_extension == "doc":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
            elif file_extension == "txt":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "text/plain"
            elif file_extension == "csv":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "text/csv"
            elif file_extension == "xls":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = (
                    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
                )
            elif file_extension == "rtf":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "application/rtf"
            elif file_extension == "png":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "image/png"
            elif file_extension == "jpg":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "image/jpeg"
            elif file_extension == "jpeg":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "image/jpeg"
            elif file_extension == "gif":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "image/gif"
            elif file_extension == "bmp":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "image/bmp"
            elif file_extension == "ico":
                body = base64.b64decode(attachment["contentBytes"])
                contentType = "image/vnd.microsoft.icon"

            s3_metadata = {
                "inbox": inbox_name,
                "email_id": email_id,
                "attachment_id": attachment_id,
                "file_name": file_name,
                "date": file_date,
                "contentType": contentType,
                "url": f"https://{self.aws_config['bucket_name']}.s3.amazonaws.com/{file_date}_{file_name_with_underscores}",
            }

            self._s3_client.put_object(
                Bucket=self.aws_config["bucket_name"],
                Key=file_name,
                Body=body,
                ContentType=contentType,
                Metadata=s3_metadata,
            )

            return s3_metadata
        except Exception as e:
            print(f"Error in _store_attachments_in_s3: {e}")
            return None

    def store_conversation(self, conversation_object):
        if not self.enabled:
            print("Storage is not configured - skipping conversation storage")
            return
        emails = conversation_object["emails"]
        new_attachments = []
        for email in emails:
            attachments = email["attachments"]
            for attachment in attachments:
                new_attachment = {
                    "attachment_id": attachment["attachment_id"],
                    "file_name": attachment["file_name"],
                    "date": self.utils._format_date(attachment["date"]),
                    "contentType": attachment["contentType"],
                }
                new_attachments.append(new_attachment)
            email["attachments"] = new_attachments
        conversation_object["emails"] = emails
        if self._check_if_conversation_exists(conversation_object["conversation_id"]):
            self._conversations_collection.update_one(
                {"conversation_id": conversation_object["conversation_id"]},
                {"$set": conversation_object},
            )
            print(
                f"Updated conversation document saved in mongo: {conversation_object['conversation_id']}"
            )
        else:
            self._conversations_collection.insert_one(conversation_object)
            print(
                f"New conversation document saved in mongo: {conversation_object['conversation_id']}"
            )

    def store_conversation_and_attachments(self, conversation_object):
        if not self.enabled:
            print("Storage is not configured - skipping conversation storage")
            return
        if self._check_if_conversation_exists(conversation_object["conversation_id"]):
            print(
                f"Conversation document already exists in mongo: {conversation_object['conversation_id']}"
            )
            return
        emails = conversation_object["emails"]
        new_attachments = []
        for email in emails:
            attachments = email["attachments"]
            for attachment in attachments:
                file_metadata = self._store_attachments_in_s3(
                    conversation_object["inbox"], email["email_id"], attachment
                )
                new_attachment = {
                    "attachment_id": attachment["attachment_id"],
                    "file_name": attachment["file_name"],
                    "date": self.utils._format_date(attachment["date"]),
                    "contentType": attachment["contentType"],
                    "url": file_metadata["url"],
                }
                new_attachments.append(new_attachment)
            email["attachments"] = new_attachments
        conversation_object["emails"] = emails
        if self._check_if_conversation_exists(conversation_object["conversation_id"]):
            self._conversations_collection.update_one(
                {"conversation_id": conversation_object["conversation_id"]},
                {"$set": conversation_object},
            )
            print(
                f"Updated conversation document saved in mongo: {conversation_object['conversation_id']}"
            )
        else:
            self._conversations_collection.insert_one(conversation_object)
            print(
                f"New conversation document saved in mongo: {conversation_object['conversation_id']}"
            )



import requests
import json
from datetime import timedelta
from yougotmail._utils._utils import Utils
from datetime import datetime, timezone

utils = Utils()


class MSWebhook:
    def __init__(self, client_id, client_secret, tenant_id):
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        self.utils = Utils()

    def create_microsoft_graph_webhook(
        self, inbox: str, api_url: str, client_state: str
    ):
        """
        Creates a Microsoft Graph webhook subscription for inbox messages
        """

        access_token = self.utils._generate_MS_graph_token(
            client_id=self.client_id,
            client_secret=self.client_secret,
            tenant_id=self.tenant_id,
        )

        if not access_token:
            print("Error: MICROSOFT_GRAPH_ACCESS_TOKEN environment variable not set")
            return None

        url = "https://graph.microsoft.com/v1.0/subscriptions"

        headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json",
        }

        # Calculate expiration datetime (max 3 days for mail resources)
        expiration_time = utils._now_utc() + timedelta(
            days=2, hours=23
        )  # Just under 3 days
        expiration_iso = expiration_time.strftime("%Y-%m-%dT%H:%M:%S.0000000Z")

        payload = {
            "changeType": "created",
            "notificationUrl": api_url,
            "resource": f"/users/{inbox}/mailfolders('inbox')/messages",
            "expirationDateTime": expiration_iso,
            "clientState": client_state,
        }

        try:
            print(f"Making POST request to: {url}")
            print(f"Payload: {json.dumps(payload, indent=2)}")

            response = requests.post(url, headers=headers, json=payload)

            print(f"Status Code: {response.status_code}")
            print(f"Response Headers: {dict(response.headers)}")

            if response.status_code in [200, 201]:
                subscription_data = response.json()
                print("Webhook subscription created successfully!")
                print(f"Subscription ID: {subscription_data.get('id')}")
                print(f"Response: {json.dumps(subscription_data, indent=2)}")
                return subscription_data
            else:
                print("Error creating webhook subscription:")
                print(f"Status: {response.status_code}")
                print(f"Response: {response.text}")
                return None

        except requests.exceptions.RequestException as e:
            print(f"Request failed: {e}")
            return None
        except json.JSONDecodeError as e:
            print(f"JSON decode error: {e}")
            print(f"Response text: {response.text}")
            return None

    def get_active_subscriptions_for_inbox(self, inbox):
        token = self.utils._generate_MS_graph_token(
            client_id=self.client_id,
            client_secret=self.client_secret,
            tenant_id=self.tenant_id,
        )

        headers = {"Authorization": f"Bearer {token}"}

        response = requests.get(
            "https://graph.microsoft.com/v1.0/subscriptions", headers=headers
        )
        response.raise_for_status()

        subscriptions = response.json().get("value", [])

        matching_subs = [
            sub
            for sub in subscriptions
            if f"/users/{inbox.lower()}/mailfolders('inbox')/messages"
            in sub["resource"].lower()
        ]

        result = {
            "total_subscriptions": len(matching_subs),
            "inbox": inbox,
            "subscriptions": [
                {
                    "id": sub["id"],
                    "expiration_date_time": sub["expirationDateTime"],
                    "notification_url": sub["notificationUrl"],
                }
                for sub in matching_subs
            ],
        }

        return result

    def renew_subscriptions(self, target_inbox):
        token = self.utils._generate_MS_graph_token(
            client_id=self.client_id,
            client_secret=self.client_secret,
            tenant_id=self.tenant_id,
        )
        headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }

        # Get current subscriptions
        r = requests.get(
            "https://graph.microsoft.com/v1.0/subscriptions", headers=headers
        )
        r.raise_for_status()
        subscriptions = r.json().get("value", [])

        renewals = 0

        for sub in subscriptions:
            sub_id = sub["id"]
            extracted_inbox = (
                sub["resource"]
                .split("/users/")[1]
                .split("/mailfolders('inbox')/messages")[0]
            )
            if extracted_inbox.lower() != target_inbox.lower():
                continue

            expiration = datetime.fromisoformat(
                sub["expirationDateTime"].replace("Z", "+00:00")
            )

            # Renew if expiring within 36h
            if expiration - datetime.now(timezone.utc) < timedelta(hours=36):
                # Format the expiration time exactly as Microsoft expects it
                new_expiration = (
                    datetime.now(timezone.utc) + timedelta(days=2, hours=23)
                ).strftime("%Y-%m-%dT%H:%M:%S.0000000Z")
                print(f"🔁 Renewing subscription {sub_id} (expires at {expiration})")

                patch_body = {"expirationDateTime": new_expiration}
                patch = requests.patch(
                    f"https://graph.microsoft.com/v1.0/subscriptions/{sub_id}",
                    headers=headers,
                    json=patch_body,
                )
                print(f"🔧 Renewal response: {patch.status_code} - {patch.text}")
                if patch.status_code == 200:
                    renewals += 1

        return {
            "status": "done",
            "subscriptions_checked": len(subscriptions),
            "inbox": target_inbox,
            "message": (
                f"renewed {renewals} subscriptions"
                if renewals > 0
                else "no subscriptions renewed"
            ),
        }

    def delete_subscription(self, subscription_id):
        url = f"https://graph.microsoft.com/v1.0/subscriptions/{subscription_id}"
        headers = {
            "Authorization": f"Bearer {self.utils._generate_MS_graph_token(self.client_id, self.client_secret, self.tenant_id)}",
            "Content-Type": "application/json",
        }
        response = requests.delete(url, headers=headers)
        return response.json()



import requests
import datetime
from dateutil.parser import parse as parse_date
from yougotmail._utils._utils import Utils
from yougotmail.retrieve.retrieve_emails import RetrieveEmails
from yougotmail.retrieve.retrieve_attachments import RetrieveAttachments
from yougotmail.retrieve.retrieval_utils import RetrievalUtils


class RetrieveConversations:
    def __init__(
        self,
        client_id,
        client_secret,
        tenant_id,
        mongo_url="",
        mongo_db_name="",
        email_collection="",
        conversation_collection="",
        attachment_collection="",
        aws_access_key_id="",
        aws_secret_access_key="",
        region_name="",
        bucket_name="",
    ):
        self.utils = Utils()
        self.token = self.utils._generate_MS_graph_token(
            client_id, client_secret, tenant_id
        )

        # Store storage configuration but don't initialize yet
        self._storage_config = {
            "mongo_url": mongo_url,
            "mongo_db_name": mongo_db_name,
            "email_collection": email_collection,
            "conversation_collection": conversation_collection,
            "attachment_collection": attachment_collection,
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
            "region_name": region_name,
            "bucket_name": bucket_name,
        }
        self._db_storage = None

        # Initialize other components with the same config
        self.retrieve_emails = RetrieveEmails(
            client_id, client_secret, tenant_id, **self._storage_config
        )
        self.retrieve_attachments = RetrieveAttachments(
            client_id, client_secret, tenant_id, **self._storage_config
        )
        self.retrieval_utils = RetrievalUtils(client_id, client_secret, tenant_id)

    def _ensure_storage(self):
        """Lazy initialization of storage"""
        if self._db_storage is None:
            from yougotmail.storage.storage import Storage

            self._db_storage = Storage(**self._storage_config)
        return self._db_storage

    @property
    def db_storage(self):
        """Property to access storage only when needed"""
        return self._ensure_storage()

    def get_conversation(
        self,
        inbox="",
        conversation_id="",
        range="last_365_days",
        start_date="",
        start_time="",
        end_date="",
        end_time="",
        subject="",
        sender_name="",
        sender_address="",
        read="all",
        attachments=False,
        storage=None,
    ):
        try:
            if inbox == "":
                raise Exception("Inbox is required")
            if conversation_id != "":
                return self._get_conversation_by_conversation_id(
                    inbox, conversation_id, attachments, storage
                )

            # get all messages fitting criteria
            emails_matching_conversation_filters = (
                self.retrieve_emails._get_emails_from_one_inbox_without_print(
                    inbox=inbox,
                    range=range,
                    start_date=start_date,
                    start_time=start_time,
                    end_date=end_date,
                    end_time=end_time,
                    subject=[subject],
                    sender_name=[sender_name],
                    sender_address=[sender_address],
                    recipients=[],
                    cc=[],
                    bcc=[],
                    folder_path=[],
                    drafts=False,
                    archived=False,
                    deleted=False,
                    sent="all",
                    read=read,
                    attachments=attachments,
                )
            )

            if emails_matching_conversation_filters is None:
                print(f"No conversation found in inbox {inbox} for the given filters")
                return None
            emails_matching_conversation_filters = (
                emails_matching_conversation_filters.get("emails")
            )
            conversation_ids = set(
                [
                    email.get("conversation_id")
                    for email in emails_matching_conversation_filters
                ]
            )

            if len(conversation_ids) == 1:
                return self._get_conversation_by_conversation_id(
                    inbox, list(conversation_ids)[0], attachments, storage
                )
            elif len(conversation_ids) > 1:
                # Create a mapping of conversation IDs to their details
                conversation_details = []
                for email in emails_matching_conversation_filters:
                    conversation_details.append(
                        {
                            "conversation_id": email.get("conversation_id"),
                            "subject": email.get("subject"),
                            "sender": email.get("sender_address"),
                        }
                    )
                # Remove duplicates by converting to dictionary and back to list
                unique_conversations = list(
                    {v["conversation_id"]: v for v in conversation_details}.values()
                )
                number_of_conversations = len(unique_conversations)
                print(
                    f"More than one conversation found in inbox {inbox} for the given filters."
                )

                print(
                    f"Found {number_of_conversations} conversations. Narrow down your search by providing more filters or using the conversation_id parameter."
                )
                print(f"Conversations found:")
                conversation_count = 0
                for conv in unique_conversations:
                    conversation_count += 1
                    print(
                        f"{conversation_count}. Conversation ID: {conv['conversation_id']}"
                    )
                    print(f"   Subject: {conv['subject']}")
                    print(f"   Sender: {conv['sender']}")
                    print("--------------------------------")
                return None
            else:
                print(f"No conversation found in inbox {inbox} for the given filters")
                return None
        except Exception as e:
            print(f"Error in get_conversation: {e}")
            return None

    def _re_structure_email_for_conversation(self, email, inbox, attachment_list):
        unique_body = email.get("uniqueBody", {}).get("content", "")
        if unique_body == "":
            body = email.get("body", {}).get("content", "")
        else:
            body = unique_body
        body = self.retrieval_utils._trim_reply_section(body)

        from_field = email.get("from", {}).get("emailAddress", {})
        sender_name = from_field.get("name", "")
        sender_address = from_field.get("address", "")

        return {
            "received_date": email["receivedDateTime"],
            "folder_name": self.retrieval_utils._get_folder_name(
                email["parentFolderId"], inbox
            ),
            "sender_name": sender_name,
            "sender_address": sender_address,
            "recipients": [
                {
                    "recipient_name": recipient["emailAddress"]["name"],
                    "recipient_address": recipient["emailAddress"]["address"],
                }
                for recipient in email["toRecipients"]
            ],
            "cc": [
                {
                    "cc_recipient_name": cc_recipient["emailAddress"]["name"],
                    "cc_recipient_address": cc_recipient["emailAddress"]["address"],
                }
                for cc_recipient in email["ccRecipients"]
            ],
            "bcc": [
                {
                    "bcc_recipient_name": bcc_recipient["emailAddress"]["name"],
                    "bcc_recipient_address": bcc_recipient["emailAddress"]["address"],
                }
                for bcc_recipient in email["bccRecipients"]
            ],
            "subject": email["subject"],
            "body": body,
            "attachments": attachment_list,
        }

    def _get_conversation_by_conversation_id(
        self, inbox, conversation_id, attachments, storage
    ):
        try:
            conversation_id_without_prefix = conversation_id.split("_")[1]
            url_filter = (
                f"$filter=conversationId eq '{conversation_id_without_prefix}'&"
                f"isDraft eq false&"
                f"$select=from,conversationId,hasAttachments,receivedDateTime,subject,toRecipients,ccRecipients,bccRecipients,uniqueBody,body,parentFolderId"
            )
            url = (
                f"https://graph.microsoft.com/v1.0/users/{inbox}/messages?{url_filter}"
            )
            headers = {
                "Authorization": f"Bearer {self.token}",
                "Accept": "application/json",
                "Prefer": 'outlook.body-content-type="text"',
            }
            try:
                emails_list = []
                page_count = 0
                email_count = 0
                while url and page_count < 100000:
                    page_count += 1
                    # print(f"Getting page - {page_count}")
                    response = requests.get(url, headers=headers)
                    response.raise_for_status()
                    data = response.json()
                    # print(data)
                    emails = data["value"]
                    next_page_link = data.get("@odata.nextLink")

                    for email in emails:
                        if not attachments:
                            attachment_list = []
                        elif attachments:
                            attachment_list = self.retrieve_attachments._get_attachments_from_one_email(
                                inbox, email
                            )
                        re_structured_email = self._re_structure_email_for_conversation(
                            email, inbox, attachment_list
                        )
                        filtered_email = self.retrieval_utils._filter_email_outputs(
                            re_structured_email,
                            archived=False,
                            deleted=False,
                            sent="all",
                        )
                        if filtered_email is not None:
                            email_count += 1
                            print(f"\nEmail {email_count} in {inbox}:")
                            print(f"  Inbox:     {inbox}")
                            print(
                                f"  From:      {filtered_email.get('sender_address')}"
                            )
                            print(f"  Subject:   {filtered_email.get('subject')}")
                            print(f"  Date:      {filtered_email.get('received_date')}")
                            print(f"  Folder:    {filtered_email.get('folder_name')}")
                            print(
                                f"  Attachments: {len(filtered_email.get('attachments', []))}"
                            )
                            emails_list.append(filtered_email)
                    url = next_page_link

            except requests.exceptions.RequestException as e:
                print(f"Error making request: {e}")
                return None

            if len(emails_list) == 0:
                print(f"No emails found in inbox {inbox} for the given filters")
                return None

            # Sort emails by received_date, most recent first
            def safe_parse_date(email):
                try:
                    received_date = email.get("received_date")
                    if received_date:
                        return parse_date(received_date)
                    else:
                        return (
                            datetime.datetime.min
                        )  # Put emails with no date at the end
                except Exception:
                    return datetime.datetime.min  # Put unparseable dates at the end

            emails_list.sort(key=safe_parse_date, reverse=True)

            conversation_object = {
                "inbox": inbox,
                "conversation_id": conversation_id,
                "number_of_emails_found": len(emails_list),
                "emails": emails_list,
            }

            if storage in ["emails", "emails_and_attachments"]:
                storage_instance = self._ensure_storage()
                if storage == "emails":
                    storage_instance.store_conversation(conversation_object)
                else:
                    storage_instance.store_conversation_and_attachments(
                        conversation_object
                    )
                conversation_object = self.utils._convert_datetimes(
                    self.utils._remove_objectid_from_list(conversation_object)
                )

            return conversation_object
        except Exception as e:
            print(f"Error in get_conversation_by_conversation_id: {e}")
            return None
